	PAGE	82,132
	TITLE	EGADDRV -- EGAD Screen Print (c) 1993 LS Software
	SUBTTL	EGAD Printer Driver
;
;	Copyright (C) 1988, 1989, 1990, 1991, 1992, 1993
;		LS Software
;		8139 E. Mawson
;		Mesa, AZ 85207
;		(602) 380-9175
;
; PURPOSE
; =======
;
;	EGAD_DRV is a FORTRAN callable subroutine to print the contents
;	of the screen to one of the supported printers.
;
; FORTRAN INTERFACE
; =================
;
;	CALL EGAD_DRV(COLORS,CONFIG,ERRCOD)
;
; INPUT	VARIABLES
; ===============
;	INTEGER*2 COLORS(16)	Screen color mapping values
;	ARRAY     CONFIG	CONFIG_DATA structure, as defined in GLOBAL.INC.
;
; OUTPUT VARIABLES
; ================
;	INTEGER*2 ERRCOD    Zero if no error, else error code
;
; GLOBAL VARIABLES (BIOS maintained)
; ==================================
;	BYTE	  0040:0049H Current screen mode
;	BYTE	  0040:004AH Number of columns being displayed
;	BYTE	  0040:0062H Video page	now on screen
;	BYTE	  0040:0084H Number of rows-1 being displayed (EGA only)
;
; LOCAL	VARIABLES
; ===============
;  in EGADDATA.ASM:
;	GMAP0..GMAPx	ARRAY	Several "maps", i.e., translations, which
;				are used to translate screen pixels to
;				printer pixels.  Each takes into account
;				the aspect ratio of a single screen mode,
;				and matches that to a specific printer
;				graphics mode.
;	COLOR_XLAT	BYTE ARRAY	
;				Maps abstract printer color numbers to
;				physical printer colors.
;	GMODES		WORD ARRAY
;				One record for each printer graphics mode.
;				Each contains a pointer to the string to send
;				to make the printer enter that mode, plus
;				the printer resolution (X,Y dots/inch).
;	PMODE_PRTn	STRING	Printer graphics control string sent to
;				initiate graphics printing.  This string
;				may contain a length field, and will be
;				followed by the actual graphics data.
;	PMODE_LENn	STRING or WORD
;				The length field within the printer control.
;	GR_OFF		STRING	This printer control string is sent after
;				the actual data is sent to the printer.
;				Also sent between each "ribbon" of color,
;				it also does a carriage return if required.
;	CRLF		STRING	Carriage return & linefeed pair
;	GSETUP			If GSET_PER_MODE is defined, this is a table
;				of pointers to setup strings, one for each
;				printer mode.  Otherwise this is the setup
;				string itself, and is used for all printer
;				modes.
;	NSPACE		STRING	Sent after all graphics printing is complete.
;				Should restore line spacing, exit graphics
;				mode, etc.
;	GR_YELLOW,
;	GR_MAGENTA,
;	GR_CYAN,
;	GR_BLACK	STRING	Each of these strings is sent to prepare the
;				printer to receive graphics of that color.
;	RIBBON_RESET	STRING	Sent after the printout is complete.
;	PATTERN_x	BYTE ARRAY
;				The byte previous to each PATTERN_x label
;				contains the number of data bits used
;				(8 for most printers).  The byte at the
;				label position contains the number of bytes
;				following.  The following bytes contain
;				bit patterns used for mapping screen pixels
;				to printer pixels in sequence.
;	DITHER_LEVELS	EQUATE	The number of bytes in each dither pattern
;	DITHER_WIDTH	EQUATE	The number of dither patterns ("colors")
;	NDITHERS	EQUATE	Number of dither tables
;	DITHER_TABLE	BYTE ARRAY
;				An array of size DITHER_LEVELS * DITHER_WIDTH *
;				NDITHERS, these bytes contain the dithering
;				patterns used to make gray shades and color
;				levels.
;  in this file:
;	CFG		WORD ARRAY
;				Copy of configuration parameters
;	COLOR_TABLE	WORD ARRAY
;				Screen color to printer color mapping table
;	GMAPNUM		BYTE	Graphics setup being used
;	GMODE		BYTE	Printer graphics mode # to use
;	ROTATE		BYTE	Nonzero if image rotated
;	LASER_RPT	BYTE	Number of (horizontal) scan lines to produce
;				for each pixel on screen. [Not used on matrix
;				printers.]
;	ERRCOD		WORD	Temporary storage for return error code
;	SCREEN_LINE	WORD	Screen line (row) being processed
;	SCREEN_COL	WORD	Screen column being processed
;	GRAPH_LMARG	WORD	Printout left margin (in printer units,
;				columns or pixels)
;	GRAPH_TMARG	WORD	Printout top margin (rows, pixels)
;	BITS_PTR	WORD	Pointer to current pixel-set pattern
;	PRN_REPEAT	BYTE	Printer pixels per screen pixel
;	OUT_LOC		WORD	Byte offset into printer graphics output buffer
;	OUT_BITS	BYTE	Print pixel bit positions
;	ML_NUM		BYTE	Microline number currently being produced
;				(0..3, for Xerox 4020 only)
;	BUF_START	WORD	Pointer to start of current graphics buffer
;	DITHER_PTR	WORD	Pointer to current dither table
;	DITHER_NUM	BYTE	Dither table number (0..NDITHERS-1)
;	LAST_COMPRESSION WORD	Pointer to string sent for prior data-
;				compression mode (only for devices which
;				support data-compression)
;	ML_BUF		BYTE ARRAY
;				Printer graphics output data area
;				Subdivided into ML_YW, ML_MG, ML_CY,
;				ML_ENCODE, and ML_BK on color devices.
;
; OTHER	SUBROUTINES CALLED
; ========================
;	GETMODE	 Call BIOS to get screen mode
;	READPT	 Read graphics point from screen
;	BEEP	 Audible alert (process complete)
;	TEST_KBD See if key pressed
;	READ_KBD Read key from keyboard
;	INT 14H, MS-DOS	serial printer service
;	INT 17H, MS-DOS	printer	output
;
;

	PUBLIC	EGAD_DRV, BOX_BUF, FLAG, FLAG1, SIGNATURE, TSR
	PUBLIC	PALETTE_2,PALETTE_4,PALETTE_16,PALETTE_TEXT

	PUBLIC	PRINTER_MSG, HDR_VERSION, HDR_NVARS
	PUBLIC	INST_MCB

	PUBLIC	PRINT_ONE_CHAR		; FOR TESTING ONLY!!!

	SUBTTL	Segment & BIOS definitions
	PAGE +

	INCLUDE	MACROS.INC
	INCLUDE	GLOBAL.INC		; Define segments, &c.
	INCLUDE VERSION.INC		; EGAD Version information

_DATA		SEGMENT WORD PUBLIC 'DATA'
	EXTRN	DISP_COLS:WORD, DISP_ROWS:WORD
	EXTRN	BYTES_ROW:WORD
	EXTRN	VIDEO_MODE:BYTE, VIDEO_PAGE:BYTE
	EXTRN	VMODE_TYPE:WORD
_DATA		ENDS

	include	manufact.inc

	ASSUME	CS:_TEXT, DS:_DATA, ES:_DATA, SS:STACK

	EXTRN	READPT:PROC
	EXTRN	START:PROC
	EXTRN	GETMODE:PROC, GET_MODE_PTR:PROC, BEEP:PROC
	EXTRN	TEST_KBD:PROC, READ_KBD:PROC

	FALSE   EQU     0
	TRUE    EQU     NOT FALSE

P_COLORS	EQU <DWORD PTR [BP+14]>
P_CONFIG	EQU <DWORD PTR [BP+10]>
P_ERRCOD	EQU <DWORD PTR [BP+06]>

TEXT_WID	EQU 80		; For now a max. of 80 columns text printout
IFDEF WIDE
 P_WID = 132
ELSE
 P_WID = 80
ENDIF

	SUBTTL	Printer Specific Data area
	PAGE +
INCLUDE EGADDATA.ASM

	SUBTTL	Driver Data Area
	PAGE +

_DATA	SEGMENT

CFG	CONFIG_DATA <>		; Copy of passed-in configuration structure

COLOR_TABLE LABEL	BYTE
		DB	16 DUP (?)

;	Driver Mode variables: Set once for each call of the driver

MONOCH_TEXT	DB	?	; TRUE if in monochrome text mode
GMAPNUM		DB	?	; Graphics mode setup being used
GMODE		DB	?	; Printer graphic mode being used, index to GMODES
ROTATE		DB	?	; Is picture being rotated?

;	Driver mode variables for Laser & Inkjet printers only:

IF (PRN_BYTE EQ BYTE_HORIZ)
 LASER_RPT	DB	?	; Number of horiz. lines per screen pixel
ENDIF

;	Driver temporary variables: change as the driver processes the screen.

		EVEN

; SCREEN_* are used by the graphics section only.
; POS_*    are used by alpha section only.
; These two sets of variables are aliased to save a few bytes.
SCREEN_LINE	LABEL	WORD	; Screen line being processed
POS_NOW		LABEL	WORD	; Current reading position
 NOW_COL	DB	?	;   -- these are 'union'ized with POS_NOW
 NOW_ROW	DB	?
SCREEN_COL	LABEL	WORD	; current pixel on line
POS_SAVED	LABEL	WORD	; Saved cursor position
 SAVE_COL	DB	?	;   -- 'union'ized with POS_SAVED
 SAVE_ROW	DB	?

GRAPH_LMARG	DW	?	; Left margin (in printer columns/pixels)

BITS_PTR	DW	?	; Pointer to current pixel pattern
PRN_REPEAT	DW	?	; Printer pixels per screen pixel
				; ("Repeat" factor from printer table)
OUT_LOC		DW	?	; Byte offset into printer graphics buffer
OUT_BITS	DB	?	; Print	pixel bit positions

IF ((prn_typ AND typ_mat24) NE 0)
 ; These are used to compose the several 8-bit high partial commands
 ; for 24-pin matrix printers.
 MAX_BUFLEN	DW	?	; Size of largest buffer in print group
 BUF_NUM	DB	?	; Number (0..PIN24-1) of current buffer
ENDIF
IF (prn_mfg eq mfg_xerox)	; Xerox 4020: strange duck
 ; This printer requires us to assemble some number of microlines into
 ; 1 printer command
 ML_NUM		DB	?
ENDIF

ERRCOD		DW	?	; Return error code

BUF_START	DW	ML_BUF+2 ; Start of current graphics buffer.
				; On 24-pin printers ONLY, this varies
				; as we iterate thru the pin groups.
				; Fixed for all other printers.
DITHER_PTR	DW	?	; Pointer to current dither table
DITHER_NUM	DB	0	; Dither table number.
				; Counts from 0..(NDITHERS-1)
IFDEF xmit_nochg
 LAST_COMPRESSION DW	0	; Pointer to string to send for determined
				; compression value
ENDIF
	EVEN
;
;	Printer Buffers -- used while building print lines
;
;	For B&W printers, we re-use the crop box buffer.
;	On color printers, that won't not be big enough, so we
;	declare new buffers here.

BOX_BUF LABEL	BYTE		; Area to save crop box data

ML_BUF	LABEL	BYTE

	SHOWIT	<BUFLEN>,%(BUFLEN)

IF (((prn_typ AND typ_color) NE 0) AND (prn_mfg NE mfg_adobe))
 ; For all color printers EXCEPT PostScript...
 ; Define three or four color planes.  Note that these could actually
 ; be any color, and not necessarily in this order, depending on how
 ; the printer defines color.
 ML_YW	DB	BUFLEN DUP (?)	; Yellow buffer
 IFDEF PIN24
	DB	BUFLEN*(PIN24-1) DUP (?) ; 2 or 3 additional buffers
 ENDIF
 ML_MG	DB	BUFLEN DUP (?)	; Magenta buffer
 IFDEF PIN24
	DB	BUFLEN*(PIN24-1) DUP (?)
 ENDIF
 ML_CY	DB	BUFLEN DUP (?)	; Cyan buffer
 IFDEF PIN24
	DB	BUFLEN*(PIN24-1) DUP (?)
 ENDIF

  ; Always define a black buffer.
  IF (NBUFFERS EQ 4)
    ML_BK DB	BUFLEN DUP (?)	; Black color plane buffer
   IFDEF PIN24
	DB	BUFLEN*(PIN24-1) DUP (?)
   ENDIF
  ELSE
    ; Printer does not have black color plane for graphics. Create a buffer
    ; for black text (if color text), or for all text (otherwise).
    IF (COLOR_TEXT)
      ; Black buffer must follow the 3d color buffer in sequence.
      ; Requires enough space for TEXT_WID characters plus as many attributes.
      ML_BK DB (TEXT_WID * 2) DUP (?)
    ELSE
      ; All text uses the black buffer, so use the yellow buffer.
      ML_BK EQU ML_YW
    ENDIF
  ENDIF
  ; While the color plane buffers must be sequential, the encoding
  ; buffer can follow.
  IF (PRN_GR AND GR_RUN) NE 0
    ; Temporary buffer for run-length encoding.
    ML_ENCODE DB BUFLEN DUP (?)
  ENDIF

 ; Mark end of all graphics buffers
 ML_END	EQU	$
 ML_WORDS EQU	(ML_END-ML_BUF)/2 ; Number of words in printer buffers
ELSE
 ML_BK	DB	BUFLEN DUP (?)
 IFDEF PIN24
   ML_OTHER DB	BUFLEN*(PIN24-1) DUP (?) ; 2 or 3 additional buffers
 ENDIF
 ML_END	EQU	$
 ML_WORDS EQU	(BUFLEN)/2	; # of words to clear

 NBUFFERS EQU	1

 IF (PRN_GR AND GR_RUN)
  ML_ENCODE DB	BUFLEN DUP (?)	; Temporary buffer for run-length encoding
 ENDIF

ENDIF

IF (box_buf + box_buf_siz) GT ($)
	db	((box_buf+box_buf_siz)-$) dup (0)
ENDIF
	EVEN

_DATA		ENDS

	SUBTTL	EGAD Printer Driver (Main Driver, EGAD_DRV)
	PAGE +

_TEXT	SEGMENT
;
;***********************  Subroutine entry point ************************
;
;	EGAD_DRV
;
;	This routine produces one graphic print out.  It does this
;	by translating the color graphics adapter memory image into
;	color print lines.
;
EGAD_DRV PROC FAR
	PUSH	BP		; Save stack starting point
	MOV	BP,SP
	PUSH	DS
	PUSH	ES
	PUSH	DI
	PUSH	SI
	MOV	AX,_DATA	; Our data segment
	MOV	DS,AX
; ********************
;	INT 3		; Debugging tool
; ********************
; Copy configuration structure to our local copy
;
	LES	BX,P_CONFIG	; Address of configuration structure
	MOV	SI,OFFSET CFG	; Address of our copy of it
	MOV	CX,CFG_LEN	; Size (in bytes) to copy
EF:	MOV	AL,ES:[BX]	;   get a byte
	MOV	[SI],AL		;   copy it
	INC	BX
	INC	SI
	LOOP	EF
;
; Check page setup (form feed, printer reset before, etc.)
;
	mov	al,byte ptr cfg.do_ff
	mov	ah,al		; Copy into AH
	and	ah,PG_FF_AFTER	; Make 1 (FF after) or 0 (no FF after)
	mov	byte ptr ff,ah	; Remember whether to eject page after
	test	al,PG_DONT_PRINT ; Don't print, just eject?
	jnz	ds_ff		;   If so, skip printing section.
 IFDEF PRINTER_RESET
	test	al,PG_RESET_BEFORE ; Reset page before beginning?
	jz	DM_1		; If not, skip next:

	mov	si,offset PRINTER_RESET
	call	print_it
	jnz	ds_err
 ENDIF
;
; Get current video mode
;
DM_1:	CALL	GETMODE

	call	get_mode_ptr

IFDEF DEBUG
	call	showmode
ENDIF

	cmp	ch,MAX_MODE	; Unsupported mode?
	ja	DS_ERR

	and	cl,vm_mono	; Save Monochrome-text-mode flag
	mov	monoch_text,cl
;
; Copy color array to local storage
;
	mov	cl,vmode_type.vmode_maxcolor[bx]
	and	cl,15		; Max color table size is 16
	sub	ch,ch
	MOV	SI,OFFSET COLOR_TABLE

	mov	dx,1		; Destination increment
	test	cfg.tsrflags,TSRF_REVCOLOR
	jz	copycolors
	neg	dx		; Reverse order of colors
	add	si,cx		; Start at end of table
copycolors:
	inc	cx		; Number of elements in	color array
	LES	BX,P_COLORS	; Address of colors array
colorloop:
	MOV	AL,ES:[BX]	; Make a local copy of this array
	MOV	[SI],AL		; Save abstract color #
	INC	BX		; Convert Words -> Bytes
	INC	BX
	add	si,dx
	LOOP	colorloop

IFDEF DEBUG
	call	showpalette
ENDIF

	MOV	AX,DS		; DS -> ES
	MOV	ES,AX
	CLD			; Insure string	instruction incr.

; Branch to the	correct	alpha or graphics mode
	call	get_mode_ptr	; BX -> vmode_type table entry
				; CH = current mode #
				; CL = vmode_flags byte
	; Now retrieve aspect ratio (gfx_map type):
	mov	al,vmode_type[bx].vmode_palette
	; and	al,vm_pal_aspect	; Shifts mask this off...
	shr	al,1
	shr	al,1
	shr	al,1
	shr	al,1
	cbw			; Clear AH
	mov	bx,offset DO_ALPHA
	test	cl,vm_text	; Is it a text mode?
	jnz	go_process	; Yes -- use text process
	mov	bx,offset DO_GRAPHICS ; Use graphics process

go_process:
	; Requires AL = Gfx_map number (0..4)
	CALL	BX		; Go to selected process

	mov	ax,errcod	; stop if error
	or	ax,ax
	jnz	ds_err

IFDEF BLACK
	; Reset to black ribbon
	mov	si,offset BLACK
	call	print_it
	jnz	ds_err
ENDIF
IFDEF TEXT_ATTRS
  IFNDEF NO_RESET_ATTRS
	; Reset printer to normal text
	mov	si,word ptr text_attrs	; Address of "Normal" string
	call	print_it
	jnz	ds_err
  ENDIF
ENDIF
	; Send form feed if required.
DS_FF:	mov	si,offset FF	; If no	FF, length to print is zero
	call	print_it
	jnz	ds_err

	CALL	BEEP		; Beep -- done!

DS_ERR:
	LES	BX,P_ERRCOD	; Return error code
	MOV	AX,ERRCOD
	MOV	ES:[BX],AX
	POP	SI
	POP	DI
	POP	ES
	POP	DS
	POP	BP
	RETF	3*4		; Clear	the stack (3 arguments, 4 bytes each)


IFDEF DEBUG
showpalette:
	PUSH	AX
	PUSH	BX
	PUSH	CX
	PUSH	DX
	 PUSH	AX
	  MOV	AL,'P'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'a'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'l'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'e'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'t'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'t'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'e'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'='
	  CALL	NEAR PTR PRINT_ONE_CHAR
	 POP	AX

	 MOV	SI,OFFSET COLOR_TABLE
	 mov	cx,16
pp0:	 lodsb
	 push	cx
	  ADD	AL,'0'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	 pop cx
	 loop	pp0

	 MOV	AL,13
	 CALL	NEAR PTR PRINT_ONE_CHAR
	 MOV	AL,10
	 CALL	NEAR PTR PRINT_ONE_CHAR
	POP	DX
	POP	CX
	POP	BX
	POP	AX

	RETN

showmode:
	PUSH	AX
	PUSH	BX
	PUSH	CX
	PUSH	DX
	 PUSH	AX
	  MOV	AL,'M'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'o'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'d'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'e'
	  CALL	NEAR PTR PRINT_ONE_CHAR
	  MOV	AL,'='
	  CALL	NEAR PTR PRINT_ONE_CHAR
	 POP	AX

	 mov	al,ch			; CH = mode #
	 PUSH	AX
	  SHR	AL,1			; OUTPUT HIGH NYBBLE
	  SHR	AL,1
	  SHR	AL,1
	  SHR	AL,1
	  ADD	AL,'0'
	  cmp	al,'9'
	  jle	gm_hx
	  add	al,7			; MAKE 10 -> 'A'
gm_hx:
	  CALL	NEAR PTR PRINT_ONE_CHAR
	 POP	AX

	 AND	AL,0FH			; OUTPUT LOW NYBBLE
	 ADD	AL,'0'
	 cmp	al,'9'
	 jle	gm_hy
	 add	al,7			; MAKE 10 -> 'A'
gm_hy:
	 CALL	NEAR PTR PRINT_ONE_CHAR

	 MOV	AL,13
	 CALL	NEAR PTR PRINT_ONE_CHAR
	 MOV	AL,10
	 CALL	NEAR PTR PRINT_ONE_CHAR
	POP	DX
	POP	CX
	POP	BX
	POP	AX
	retn
ENDIF




EGAD_DRV	ENDP

	SUBTTL	EGAD Printer Driver (Alpha/Text Mode Printing)
	PAGE +
;
; ******************************************************************
;
;	A L P H A    T E X T   P R I N T    S E C T I O N
;
;	This procedure handles all text modes.
;
; We have to read each character on the current line, translate its screen
; color to print color, and put the character in all the right buffers
; (black/cyan/magenta/yellow).  Then remove trailing spaces from each of
; the buffers; for each nonempty buffer, select that color ribbon & send
; the line.  We ignore background colors, as the printer won't do background
; in text mode.
;
; ******************************************************************
;
DO_ALPHA PROC NEAR

	MOV	SI,OFFSET CRLF
	CALL	PRINT_IT
	JZ	DA_0
	RETN
DA_0:
	; Initialize text attributes
	mov	gmode,-1	; Always send attribute for 1st character

	; Tell printer to print extended ASCII set
 IFDEF EXT_CHARS
	mov	si,offset ext_chars	; Graphics/Greek character set.
	call	print_it
	jz	da_1
	retn
 ENDIF

DA_1:	MOV	AX,0300H	; Read cursor position
	MOV	BH,VIDEO_PAGE
	INT	VID_INT
	MOV	POS_SAVED,DX	; Save cursor position

	mov	al,byte ptr cfg.y1
	mov	now_row,al

	; Test for too-big numbers

	mov	ax,disp_cols	; This is one more than	max column #
	cmp	cfg.x2,ax
	jl	da_1a		; Got a	legal max X -- go with it.

	dec	al		; This is max col #
	mov	cfg.x2,ax

da_1a:	mov	ax,cfg.x2
	cmp	cfg.x1,ax	; Is max < min?
	jle	da_1b		; No, we're OK
	mov	cfg.x1,ax	; Yes -- set min to max

da_1b:	mov	ax,disp_rows	; Maximum line #
	cmp	cfg.y2,ax	; Too big a number?
	jle	da_1c		; No --	go ahead
	mov	cfg.y2,ax	; Yes -- use max

da_1c:	mov	ax,cfg.y2
	cmp	cfg.y1,ax	; Is max < min?
	jle	da_main_loop	; No, we're OK
	mov	cfg.y1,ax	; Yes -- set min to max
DA_MAIN_LOOP:			; Restart each line
	mov	di,offset ml_buf ; Zero out the four buffers.
IFNDEF TEXT_ATTRS
	mov	ax,2020h	; Space characters
	mov	cx,ml_words	; Length of buffer space, in words.
	cld			; Ensure forward direction
	rep	stosw
ELSE
	add	di,2		; skip over length bytes
	mov	ax,2020h	; Space characters
	mov	cx,TEXT_WID/2	; do 80 characters
	cld			; Ensure forward direction
	rep	stosw
	xor	ax,ax		; AX = 0
	mov	cx,TEXT_WID/2	; Clear 80 bytes of attribute
	rep	stosw
ENDIF

	mov	ax,cfg.x1	; Start at this column
	mov	cx,cfg.x2	; This many columns on line (0..n)
	sub	cx,ax
	inc	cx		; Count of columns to do
	mov	now_col,al	; Starting column (cfg.x1)
	mov	di,2		; Offset into each print buffer.
				; (first 2 bytes are Length).
	cmp	cfg.centering,JUST_LEFT
	jle	da_read_loop	; 0,1 = Left justify
	cmp	cfg.centering,JUST_RIGHT
	je	da_jright	; 2 = Right justify
	; Centered text
	push	cx
	shr	cx,1		; 1/2 width
	add	di,TEXT_WID/2
	sub	di,cx
	pop	cx
	jmp	short da_read_loop

da_jright:
	add	di,TEXT_WID	; Move to right column
	sub	di,cx		; then move left by width of image

DA_READ_LOOP:
	push	di
	 mov	bh,video_page
	 mov	dx,pos_now	; Aliased over NOW_COL and NOW_ROW
	 mov	ax,0200h	; set cursor position
	 int	VID_INT
	 inc	now_col		; Move to next column
	 mov	ax,0800h	; read char & attrs
	 int	VID_INT
	pop	di

	test	monoch_text,0ffh ; Monochrome text mode?
	jz	da_read_1
	mov	ah,1		; Yes -- text color is always 1

da_read_1:
 IFDEF ASCII_XLAT
	cmp	al,80h		; Translate only chars 128..255.
	jb	da_read_1a
	sub	al,80h		; Offset into xlation table
	mov	bx,offset ascii_xlat
	xlat
 ENDIF
da_read_1a:
 IFDEF QUOTE_CHAR
	cmp	al,0		; Translate NUL's to SPACE's
 ELSE
	cmp	al,20h		; Translate < 32 (20h) to spaces.
 ENDIF
	ja	DA_READ_2
	mov	al,20h		; Space
DA_READ_2:
	xchg	ah,al		; Character in AH, Attribute in AL.
	and	al,00FH		; Foreground color only
	mov	bx,offset COLOR_TABLE
	xlat			; AL now contains printer color #

IF COLOR_TEXT
	or	al,al		; Is color 0?
	jne	da_read_nz
	mov	al,8		; Assume black

 da_read_nz:
	test	al,1		; Is there any yellow?
	jz	da_read_2m
	mov	ml_yw[di],ah	; Store character

  da_read_2m:
	test	al,2		; Any magenta?
	jz	da_read_2c
	mov	ml_mg[di],ah

  da_read_2c:
	test	al,4		; Cyan?
	jz	da_read_2b
	mov	ml_cy[di],ah

  da_read_2b:
	test	al,8		; Black?
	jz	da_read_2x
	mov	ml_bk[di],ah	; Put character in black buffer
ELSE
	mov	ml_bk[di],ah	; Put character in black buffer
	mov	ml_bk[di+TEXT_WID],al	; Save attribute byte
ENDIF

da_read_2x:
	inc	di		; Advance to next char in buffers

	loop	DA_READ_LOOP	; Read each char on the	line

; Now, for each of the four buffers, we must start at the end, looking
; for the last non-blank char.  This gives us the length of the line.
; If this is nonzero, scan back to the beginning of the buffer,
; select the appropriate ribbon and send the line.
;
;	DI points to the current buffer.
;	SI points to the proper ribbon-select code (color printers)
;
IF COLOR_TEXT
	mov	di,offset ml_buf ; FIRST buffer (yellow if color, black if not)
	mov	si,offset yellow ; Code to set printer to color for buffer
ELSE
	mov	di,offset ml_bk
ENDIF
;
;	Here, DI points to the Length word of the buffer.
;
da_print_buf_loop:

	push	di		; Save start of buffer
	  inc	di		; Skip over Include Length word
	  inc	di

	  mov	bx,TEXT_WID	; maximum screen width
	  add	di,bx		; point to last char
	  dec	di

	  mov	cx,bx		; CX = # of columns
	  mov	al,20H		; Skip SPACES.
	  std	
	  repe	scasb		; Scan backwards thru buffer for last
	  cld			; non-space, CX to be length of line.
	pop	di		; Restore DI as beginning of buffer

	JZ	DA_2		; Blank	buffer, don't send it.

	inc	cx
	mov	[di],cx		; Save length in buffer

	; Select proper ribbon:

IF COLOR_TEXT
	call	print_it	; SI points to code to set printer to
				; this color ribbon.
	push	si		; Save SI.
ENDIF
        mov	si,di		; Point SI to text buffer

	add	si,2		; Skip over length word
;	mov	gmode,-1	; Always send attribute for 1st character
;				; (moved above, wl 09/23/93)
da_pb_lp:
IFDEF TEXT_ATTRS
	mov	al,[si+TEXT_WID] ; Get attribute byte for character
	and	al,N_TEXT_ATTRS	; Ensure valid attribute
	cmp	gmode,al	; Same as last time?
	je	da_pb_go

	mov	gmode,al	; Remember attribute
	xor	ah,ah
	add	ax,ax		; Get index of attribute in AX
	PUSH	SI
	 ; Reset to normal string -- turn off all attributes
	 MOV	SI,TEXT_ATTRS
	 CALL	PRINT_IT

	 ; Turn on desired attribute
	 MOV	SI,OFFSET TEXT_ATTRS	; Table base...
	 ADD	SI,AX			; + index...
	 MOV	SI,[SI]			; Get string address from table
	 CALL	PRINT_IT
	POP	SI

da_pb_go:
ENDIF
	lodsb			; get character, increment SI
IFDEF QUOTE_CHAR
	cmp	al,20H
	jae	da_pb_1		; Doesn't need quoting if >= 32 (20h)

	push	si
	mov	si,offset QUOTE_CHAR
	call	print_it	; Send Quote string
	pop	si
da_pb_1:
ENDIF
	call	print_one_char	; Send character
	loop	da_pb_lp	; Send rest of buffer

da_pb_ok:

IF COLOR_TEXT
	mov	si,offset CR	; Send a Carriage return
	call	print_it

	pop	si		; Point SI back to ribbon-select code
da_2:	cmp	di,offset ml_bk	; Did we just print the last buffer?
				; (we're done if so.)
	je	da_3		; Yes.  Send CR/LF & go to next line.
	add	di,BUFLEN	; Go to next buffer.
	add	si,[si]		; Move to next ribbon-select code.
	inc	si
	inc	si

	jmp	da_print_buf_loop
				; Go print the next color-buffer.
ELSE
da_2:				; Come to here to skip buffer
ENDIF

DA_3:	MOV	SI,OFFSET CRLF	; End line
	CALL	PRINT_IT
	JZ	DA_4
	RETN
DA_4:	INC	NOW_ROW
	MOV	AL,NOW_ROW
	CMP	AL,byte	ptr cfg.y2  ; Last line on screen?
	JA	DA_5
	jmp	da_main_loop
DA_5:	mov	bh,video_page
	mov	dx,pos_saved
	mov	ax,0200h	; set cursor position
	int	VID_INT
IFDEF END_TEXT
	mov	si,offset END_TEXT
	call	print_it
ENDIF
	XOR	AX,AX		; Show successful completion
	RETN
DO_ALPHA ENDP

	SUBTTL	EGAD Printer Driver (Graphics Mode Printing)
	PAGE +
;
; ******************************************************************
;
;	G R A P H I C S   P R I N T   S E C T I O N
;
;	This procedure handles all graphics modes.
;
; This is structured so	that other modes can be processed by changing
; the branch table and adding the processing routines.
;
; ******************************************************************
;
;
;	Called for all graphics modes.
;	Screen->Printer map (GMAP) number in AX.
;
DO_GRAPHICS PROC NEAR

	mov	bx,cfg.enlarge	; Get enlargement factor
	or	bx,bx		; Enlargement=0? (This means "Teeny!")
	jnz	do_gr0

	; For Reduction, use smallest printing (highest numbered map)
	mov	al,hdr_secmap	; Reduction map is
	dec	al		;    last before secondary.
	inc	cfg.enlarge	; Make enlargment=1
DO_GR0:
	sub	ah,ah		; Clear AH
	mov	GMAPNUM,AL	; Remember graphics mode setup #
	mov	cl,4		; Multiply map # by 16 to get offset
	shl	ax,cl
	mov	si,hdr_map0	; Address of map 0
	add	si,ax		;   Selected map

	LODSW			; AX = X
	MOV	BX,AX
	LODSW			; AX = Y
	MOV	CX,AX
	LODSW			; AX = graphics mode
	MOV	GMODE,al
	LODSW			; AX = Repeat count
IF (PRN_BYTE EQ BYTE_HORIZ)
	MOV	LASER_RPT,al
	MOV	byte ptr PRN_REPEAT,1	; Always = 1 -- Need to use Repeat count
					; to dither into successive lines.
ELSE
	MOV	byte ptr PRN_REPEAT,al	;	(horizontal size factor)
ENDIF
	LODSW			; AX = Rotate flag
	MOV	ROTATE,al
	LODSW			; AX = Pattern
	MOV	BITS_PTR,AX	;	(vertical switch)

; Look for desired orientation
	mov	al,byte ptr cfg.orient	; Look at desired orientation
	or	al,al
	jz	DO_GR1		; 0=Auto (primary) orientation -- all set!
	dec	al		; Now AL=0 if we want portrait, 1=landscape
	cmp	al,ROTATE	; Is that the desired rotation?
	je	DO_GR1		; Yes -- accept it.
	mov	al,GMAPNUM	; Look at current map number
	cmp	al,byte ptr hdr_secmap	; Did we already choose secondary setup?
	jge	DO_GR1		; Yes -- desired orientation not available
	add	al,byte ptr hdr_secmap	; Choose secondary setup
	cmp	al,byte ptr hdr_nmap ; Ensure the mode # is legal
	jl	DO_GR0		; And choose it if legal.
				; (illegal mode #'s fall thru)
;
; **********************************************************************
;			HORIZONTAL POSITIONING
; **********************************************************************
;
;  BX is max # of printer pixels across, CX max # down.
;  Matched to the screen resolution for proper aspect ratio.
;
DO_GR1:
	mov	screen_line,cx	; Temporarily use these variables to save 
	mov	screen_col,bx	; limits of printout (in screen pixels.)
				; Used by get_printsize_rows and ..._cols.

	mov	ax,cfg.x2	; Multiply pixels in image by factor
	sub	ax,cfg.x1	; delta_x
	call	enlarge_mul
	cmp	ax,bx		; Is new delta_x <= # of horiz pixels?
	jle	dg_xok		; yes, ok.
	mov	ax,bx		; no, set to max horiz.
dg_xok:	add	ax,cfg.x1
	mov	cfg.x2,ax	; Set right side = left + width - 1

	mov	ax,cfg.y2
	sub	ax,cfg.y1	; delta_y
	call	enlarge_mul
	cmp	ax,cx		; Is new delta_y <= # of vert pixels?
	jle	dg_yok		; yes, ok.
	mov	ax,cx		; no, set to max vert.
dg_yok:	add	ax,cfg.y1
	mov	cfg.y2,ax	; Set bottom = top + height

	; Compute graphics left margin (to allow picture centering)

	cmp	cfg.centering,JUST_LEFT	; 0,1 = Left justify
	jg	lmarg_pick
	xor	ax,ax
	jmp	short lmarg_save

; ********************
; Compute left margin
; ********************
lmarg_pick:
	xor	al,al		; Get AX = width of printout, in screen coords
	call	get_print_size	; (could be screen rows or columns)

	cmp	cfg.centering,JUST_RIGHT	; 2 = Right justify
	je	lmarg_save
	shr	ax,1		; Divide by 2 for Centering

lmarg_save:

IF (PRN_BYTE EQ BYTE_HORIZ)
	; Divide by size of pattern (# of screen pixels/byte)
	; to get margin in printer bytes
	xor	dx,dx
	mov	bx,bits_ptr	; Pointer to pattern
	mov	bl,[bx]		; size of pattern
	xor	bh,bh
	div	bx		; AX = left margin in printer bytes
ELSE
	; Multiply by number of printer bytes/screen pixel
	mul	word ptr prn_repeat ; AX = left margin in printer bytes
ENDIF
	mov	graph_lmarg,ax	; Save graphic left margin

; ********************
; Determine whether to actually print
; ********************
	dec	cfg.enlarge	; Make Enlargement zero-based.
;
;   If max < min, don't print anything.
;
	mov	ax,cfg.x2
	cmp	cfg.x1,ax	; Is max < min?
	jle	dg_1b		; No, we're OK
	jmp	dg_err		; Yes, Error.

dg_1b:	mov	ax,cfg.y2
	cmp	cfg.y1,ax	; Is max < min?
	jle	dg_start	; No, we're OK
	jmp	dg_err		; Yes, Error.

; ************************************************************
;	PRINTER SETUP / INITIALIZATION.
;	Prepare for graphics printing.
; ************************************************************

DG_START:
	; Be sure we start @ left margin -- WL 08/19/94
	mov	si,offset CR
	; Begin new blank line before printout:
	;MOV	SI,OFFSET CRLF
	CALL	PRINT_IT
	JZ	DG_SETUPS
DG_START_ERR:
	JMP	DG_ERR
;
; Setup string(s)
;
DG_SETUPS:
	; GXSETUP is used to place the printer into an emulation
	; mode. On PostScript it's used to define our variables & routines.
IFDEF GXSETUP
	mov	si,offset GXSETUP
	CALL	PRINT_IT
	JNZ	DG_START_ERR
ENDIF

	; GSETUP is the graphics setup.  On some printers (HP, etc.) the setup
	; string depends on the graphic mode used (i.e., this is where the
	; resolution is selected.)
	mov	si,offset GSETUP
IFDEF GSET_PER_MODE
	; Setup string depends on graphics mode.
	; Assume SI points to table (of pointers to) setup strings.
	mov	bl,GMODE	; Get graphics mode
	xor	bh,bh
	add	bx,bx		; Use as word index
	mov	si,[si][bx]	; Get address of setup string for this mode
ENDIF
	CALL	PRINT_IT
	JNZ	DG_START_ERR

	; GSET_ALL is an additional setup string which is the same for all
	; graphic modes (e.g., to define colors for the PaintJet):
IFDEF GSET_ALL
	mov	si,offset GSET_ALL
	CALL	PRINT_IT
	JNZ	DG_START_ERR
ENDIF

; **********************************************************************
;			VERTICAL POSITIONING
; Compute top margin
; **********************************************************************
compute_top:

	cmp	cfg.vcentering,JUST_BOTTOM
	jl	print_graphics	; JUST_TOP = Use current position.

	mov	di,offset ml_buf ; Zero out the data buffers
	xor	ax,ax
	mov	cx,ml_words
	cld			; Ensure forward direction
	rep	stosw		; (CX=0 at end)

	mov	al,1		; Get AX = Height of printout, in screen coords
	mov	byte ptr ml_buf,al ; Send one byte of data (which is blank.)
	call	get_print_size	; (could be screen rows or columns)

	cmp	cfg.vcentering,JUST_BOTTOM
	je	tmarg_save	; if != , must be JUST_CENTER.
	shr	ax,1		; Divide by 2 for centering.
tmarg_save:

IF (PRN_BYTE EQ BYTE_HORIZ)
	; Multiply by number of printer scanlines per screen pixel
	mov	dx,ax		; DX = screen pixels
	mov	al,laser_rpt	; AX = printer scanlines per screen pixel
	cbw
	mul	dx		; AX = printer scanlines to use as top margin
ELSE
	; Divide by size of pattern (# of screen pixels/byte)
	; to get margin in printer bytes
	xor	dx,dx
	mov	bx,bits_ptr	; Pointer to pattern
	mov	bl,[bx]		; size of pattern
	xor	bh,bh
	div	bx		; AX = top margin in printer bytes
ENDIF
	mov	cx,ax

tmarg_loop:
	; AX is now the number of *screen* lines to leave above
	; the first *printer* line we create.
	call	print_line
	; NOTE! The PostScript driver actually NOTs the buffer each time, so we
	; re-clear it here:
	sub	al,al
	mov	byte ptr ml_buf + 2, al ; Reset first byte to zero
	loopz	tmarg_loop	; Stop if error
	jnz	dg_start_err
;
;	COMMENCE ACTUAL GRAPHICS PRINTING
;
PRINT_GRAPHICS:
	cmp	rotate,0	; Are we rotating?
	je	DG_START_NOROT

	mov	di,offset screen_col	; Column counts up.
	mov	si,offset cfg.x1
	jmp	short dg_start_go

DG_START_NOROT:
	mov	di,offset screen_line	; Line counts up.
	mov	si,offset cfg.y1

DG_START_GO:
	mov	ax,[si]		; Get X1 (or Y1)...
	mov	dx,4[si]	; ...and X2 (or Y2)
	mov	[di],ax		; Start counter at minimum value

; Reset to 1st dither table
	sub	ax,ax
	mov	dither_ptr,offset dither_table
	mov	dither_num,al	; Clear dither number
IFDEF LAST_COMPRESSION
	mov	last_compression,ax	; Clear compression
ENDIF
IFDEF ML_NUM
	mov	ml_num,al		; First microline
ENDIF
IFDEF PIN24
	; 24-pin printer special init.
	mov	MAX_BUFLEN,ax	; clear max column count
	mov	BUF_NUM,al	; clear buffer number

	; Clear all the "extra" buffers.
	push	di
	 mov	di,offset ml_buf ; Zero out the data buffers
 IFDEF NBUFFERS
	 mov	cx,(BUFLEN*PIN24*NBUFFERS)/2 ; Clear all buffers
 ELSE
	 mov	cx,(BUFLEN*PIN24)/2 ; Clear all buffers
 ENDIF
	 cld			; Ensure forward direction
	 rep	stosw		; (CX=0 at end)
	pop	di
ENDIF

	mov	si,di		; SI -> screen_line or screen_col
	mov	bx,bits_ptr	; BX -> lines_xxx_pattern

; Top of Line print loop.  Print until SCREEN_LINE>CFG.Y2 (or SCREEN_COL>CFG.X2)
; At this point:
;	BX -> pixel line pattern (BITS_PTR)
;	DX =  Maximum value for [SI] (SCREEN_LINE/SCREEN_COL depending on ROTATE)
;	SI -> counter variable (SCREEN_LINE or SCREEN_COL)
;

DG_LINES:
	call	printer_line	; Read pixels & Send out 1 printer line's worth
	jnz	dg_err		; ERROR -- bail out.

	cmp	word ptr [si],dx	; Compare counter variable to maximum.
	jle	dg_lines	; Keep looping...

DG_DONE:
IFDEF PIN24
	push	di
	 mov	di,offset ml_buf ; Clear new buffer
	 mov	ax,0
	 mov	cx,BUFLEN/2	; Clear current buffer
	 cld			; Ensure forward direction
	 rep	stosw
	pop	di
	; 24-pin printers must flush graphics buffers
	cmp	BUF_NUM,0	; If zero, buffers are flushed
	jz	DG_DONE1
	mov	bx,offset ml_bk
	mov	word ptr [bx],0
	call	PRINT_ONE_PXL_LINE	; "Print" blank line
	jmp	DG_DONE
DG_DONE1:
ENDIF

; Place	empty line at end of printout.
IFDEF MLCRLF
	MOV	SI,OFFSET MLCRLF
ELSE
	MOV	SI,OFFSET CRLF
ENDIF
	CALL	PRINT_IT
	JNZ	DG_ERR

; Exit graphics mode and reset spacing to 1/6"
	MOV	SI,OFFSET NSPACE
	CALL	PRINT_IT
	JNZ	DG_ERR

	XOR	AX,AX		; Show no error
DG_RET:	MOV	ERRCOD,AX
	RETN

DG_ERR:
;	ERRCOD is normally 0 at this point.  If the user presses ESC,
;	however, it is set to -1 -- we will clear the error and reset
;	the printer.

	inc	errcod		; Show error. (Was -1 if key pressed, make 0)
	jz	dg_done		; pressed ESC? reset printer!

	MOV	AX,1		; Got an error!
	JMP	DG_RET

; ********************
; SUBROUTINES
; ********************
; Determine size multiplied
enlarge_mul:
	inc	ax
	cmp	byte ptr cfg.enlarge,5	; 1.5x?
	je	enlarge_15
	mul	cfg.enlarge	; * factor
	dec	ax
	retn
enlarge_15:
	mov	dx,ax
	add	ax,ax		; * 2
	add	ax,dx		; * 3
	shr	ax,1		; * 1.5
	dec	ax
	retn


; Determines the size of the printout in *screen* pixels.
; Input:  AL = 0   to return printout width
;         AL = 1   to return printout height
; Output: AX =     size of printout in requested direction

get_print_size:
	cmp	rotate,al	; Equal to rotation means screen columns
	je	get_print_cols	; Not equal means screen rows
	mov	ax,screen_line	; CX = # of vertical *SCREEN* pixels down printer
	add	ax,cfg.y1
	sub	ax,cfg.y2
	jmp	short getps0

get_print_cols:
	mov	ax,screen_col	; BX is # of (horiz on screen) pixels down page
	add	ax,cfg.x1	; Subtract width of this printout
	sub	ax,cfg.x2	;  # blank pixels remaining in AX

getps0:				; for all rotations:
	cmp	ax,0
	je	getps1
	dec	ax
getps1:	retn

IF 0
	; DEBUGGING HELPER FUNCTION
foobar:
	push	cx
	mov	al,ch
	shr	al,1
	shr	al,1
	shr	al,1
	shr	al,1
	call	phex
	pop	cx

	push	cx
	mov	al,ch
	and	al,0fh
	call	phex
	pop	cx

	push	cx
	mov	al,cl
	shr	al,1
	shr	al,1
	shr	al,1
	shr	al,1
	call	phex
	pop	cx

	push	cx
	mov	al,cl
	call	phex
	pop	cx
	retn

phex:	and	al,0fh
	add	al,'0'
	cmp	al,'9'
	jle	phex1
	add	al,'a'-'9'-1
phex1:	call	print_one_char
	retn
ENDIF
DO_GRAPHICS ENDP
	SUBTTL	EGAD Printer Driver (Subroutines)
	PAGE +
;
; ******************************************************************
;
;	EGAD Printer Driver Subroutines
;
; ******************************************************************
;
;
	SUBTTL	EGAD Printer Driver (Subroutines: PRINTER_LINE)
; ************************************
;
;	PRINTER_LINE
;
;	Fills the printer buffer for one print line, then prints it.
;
;	Entry:
;		BX points to LINES_xxx_PATTERN (xxx=200,350,400),
;		which tells us how many screen lines per print row, and also
;		gives us the bit patterns to 'or' into the print buffers.
;
;		SCREEN_LINE points to the next screen line to be processed
;		if ROTATE is 0; else SCREEN_COL does.
;
;	Preserves: AX, BX, CX, DX, SI
;
PRINTER_LINE PROC NEAR

	push	ax
	push	bx
	push	cx
	push	dx
	push	si

	mov	di,offset ml_buf ; Zero out the data buffers
	xor	ax,ax
	mov	cx,ml_words
	cld			; Ensure forward direction
	rep	stosw		; (CX=0 at end)

IF (PRN_BYTE EQ BYTE_HORIZ)

	; Laser printers print bytes horizontally; thus we have
	; just one screen row/column per printer line.	

	mov	si,offset screen_line	; Variable to increment
	cmp	rotate,0
	je	pl_go
	mov	si,offset screen_col
pl_go:
	xor	ch,ch
	mov	cl,LASER_RPT

pl_loop:

; For these printers, dither tables vary by PRINT LINE.
; Determine dither table (if more than one)

  IF NDITHERS-1
	push	si
	mov	al,dither_num		; Get current dither number
	inc	al			; Go to next one
	cmp	al,NDITHERS		; If past end, go to first table
	jl	next_dither
; Go back to first dither table
	mov	si,offset dither_table
	xor	al,al
	jmp	save_dither	
next_dither:
	mov	si,dither_ptr		; Address of 1st table
	add	si,DITHER_LEVELS*DITHER_WIDTH
save_dither:	
	mov	dither_num,al
	mov	dither_ptr,si
	pop	si
  ENDIF
	call	do_pixels	; Read & process this line/column
	call	print_line	; Send processed buffers to printer

	; Clear buffers for another pass
	push	cx
	pushf
	 mov	di,offset ml_buf ; Zero out the data buffers
	 xor	ax,ax
	 mov	cx,ml_words
	 cld			; Ensure forward direction
	 rep	stosw		; (CX=0 at end)
	popf
	pop	cx

	loopz	pl_loop		; Stop when count is 0 or when Not Zero
				; (which indicates key pressed or error)

	pushf			; Save Zero flag which indicates error
	inc	word ptr [si]	; Go to next screen line or column
	popf
ELSE

	; Dot matrix printers print bytes vertically, so we have to
	; do multiple screen lines per printer line:
	xor	ch,ch
	mov	cl,[bx]		; CX = # of screen lines in one printer line
	inc	bx		; BX points to first bit pattern

	mov	dx,cfg.y2	; This is the last screen line to do.
	mov	si,offset screen_line	; Variable to increment
	cmp	rotate,0
	je	pl_loop
	mov	dx,cfg.x2	; Last column to do.
	mov	si,offset screen_col

pl_loop:
	; Multiple dithers on dot matrix are by PIXEL.
	; At beginning of line, we go back to 1st dither.
	mov	dither_ptr,offset dither_table
	mov	dither_num,0

	mov	al,[bx]		; Get the next bit pattern
	inc	bx
	mov	OUT_BITS,al
	call	do_pixels	; Process one line or column of pixels
	; Only increment screen position when we process it.
	inc	word ptr [si]	; Go to next screen line or column
	cmp	[si],dx		; Bail out if all done
	jg	pl_done
	loop	pl_loop		; Do all screen lines mapping to this print line

pl_done:
	call	print_line	; Send buffer to printer
ENDIF
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	ax

	RETN
PRINTER_LINE ENDP
	SUBTTL	EGAD Printer Driver (Subroutines: PRINT_LINE)
;
; ************************************
;
;	PRINT_LINE
;
;	Sends the printer buffers off to the printer.
;
;	Exit:	Zero Flag set if OK, clear if printer error.
;
;	Preserves:	BX, CX, DX, SI
;	Uses:		AX, DI
;
PRINT_LINE PROC	NEAR
	PUSH	SI
	PUSH	DX
	PUSH	CX
	PUSH	BX

print_line_test_kbd:
	CALL	TEST_KBD	; Get keyboard status
	jz	print_line_go	; No keys waiting

	CALL	READ_KBD
	cmp	al,27		; ESC pressed?
	jne	print_line_test_kbd	; No, read any other pending chars

	or	al,0ffh		; "Not Zero" in flags; AL = -1
	cbw			; AX = -1
	mov	errcod,ax	; ERRCOD = -1 means keyboard error
	jmp	pl_err		; stop printing

print_line_go:
;
;	Find the Lengths of the Buffers
;
	mov	bx,offset ml_buf ; Length word in first graphics buffer
	mov	cx,NBUFFERS	; number of printer buffers
	xor	al,al		; Look for zeros

do_pd0:
	xchg	dx,cx		; Save loop ctr in DX
	mov	cx,BUFLEN-2	; # data bytes
	mov	di,bx		; Set DI to last data byte in buf...
	add	di,cx		; Beginning + Data length + Length Word(2) - 1
	inc	di		; + 2 - 1 = +1 byte offset yet
	std
	repe	scasb		; Find last nonzero value
	cld
 IF (PRN_BYTE EQ BYTE_VERT)
 ; On laser printers and other printers with horizontal bytes,
 ; even blank graphics lines must be sent.  (On matrix printers
 ; we skip blank lines since we always send CR/LF anyway.)
 ; Thus, if length=0, we make length=1 here.

 ; If buffer had one byte in it, CX=0, ZF=false. 2 bytes in buffer, CX=1
 ; and ZF=false. Buffer had no bytes in it, CX=0, ZF=true.
	jz	do_pd1		; No bytes in buffer, leave length=0
 ENDIF
	inc	cx		; This is the true length
do_pd1:	mov	[bx],cx		; Save length

	add	bx,BUFLEN
	xchg	dx,cx		; DX has buffer count
	loop	do_pd0

IFDEF ML_NUM ; If separate codes for different microlines
	mov	al,ml_num	; Get current microline number
	push	ax
	 inc	ml_num		; Increment for next time
	 cmp	al,4		; Past bottom of previous microline?
	 jl	do_pd2
	 mov	si,OFFSET MLXCRLF ; Print previous microlines
	 CALL	PRINT_IT
	 sub	al,al		; Revert to 0th microline
	 mov	ml_num,al	; Save next microline number
do_pd2:	pop	ax	

	add	al,'0'		; Make '0'-'3'
	mov	si,4+OFFSET GR_BLACK
	mov	ds:[si],al	; Save code into Black command
	add	al,4		; Make '4'-'7'
	mov	si,4+OFFSET GR_MAGENTA
	mov	ds:[si],al	; Save code into magenta command
	add	al,4		; Make '8'-';'
	mov	si,4+OFFSET GR_YELLOW
	mov	ds:[si],al	; Save code into yellow command
	add	al,4		; Make '<'-'?'
	mov	si,4+OFFSET GR_CYAN
	mov	ds:[si],al	; Save code into cyan command
ENDIF

;
;	Print the buffers
;
	mov	dx,0FFFFH	; Use this to TEST against lengths to see if 0
	mov	bx,offset ml_buf ; First buffer to print

IFDEF ML_YW ; If separate color buffers are defined
	mov	si,offset gr_yellow	; Select 1st (usu. yellow) ribbon,
	call	print_a_buffer		; or graphics mode
	jnz	pl_err
	add	bx,buflen

	mov	si,offset gr_magenta	; Select 2d (usu. magenta) ribbon,
	call	print_a_buffer		; or graphics mode
	jnz	pl_err
	add	bx,buflen

	mov	si,offset gr_cyan	; Select 3d (usu. cyan) ribbon,
	call	print_a_buffer		; or graphics mode
	jnz	pl_err
	add	bx,buflen
ENDIF
;
; Process Black buffer, if printer has one.
;
IF (NBUFFERS EQ 1) OR (NBUFFERS EQ 4)
  IFDEF GR_BLACK
	test	[bx],dx		; Any data in that buffer?
	jz	pl_end		; No, skip it (& don't send Black cmd)
	mov	si,offset gr_black ; select black color
	call	print_it
	JNZ	PL_ERR
  ENDIF
	CALL	PRINT_ONE_PXL_LINE ; Print black bits
	JNZ	PL_ERR
ENDIF
pl_end:
IFNDEF PIN24
	; If required, send a CR/LF (or some other sequence) after
	; each graphics row.  (If 24-pin printer, the appropriate
	; sequences will be sent from PRINT_ONE_PXL_LINE.)
	MOV	SI,OFFSET MLCRLF
	CALL	PRINT_IT
	JNZ	PL_ERR
ENDIF

PL_OK:	XOR	AX,AX	    ; Show no error
PL_ERR:	POP	BX
	POP	CX
	POP	DX
	POP	SI
	RETN

IFDEF ML_YW ; Called by above logic for color printers
print_a_buffer:
	test	[bx],dx			; No data in buffer?
	jz	print_a_buf_exit	; Return w/o error
	call	print_it		; Select ribbon or buffer
	jnz	print_a_buf_exit	; Return if any error
	call	print_one_pxl_line	; Print buffer at [bx]
print_a_buf_exit:
	; returns "Zero" if success, "Not Zero" if error
	retn
ENDIF

PRINT_LINE ENDP
	SUBTTL	EGAD Printer Driver (Subroutines: PRINT_ONE_PXL_LINE)
IFDEF PIN24
;
; ************************************
;
;	PRINT_ONE_PXL_LINE
;
;	Sends a line of pixels to the printer.
;
;	Entry:  DS:BX	Points to data buffer to be sent to printer
;
;	Exit:	'Zero' if printed OK, not 'Zero' if error
;
;	Uses:	Preserves all
;
;	--------------------
;	24-Pin Printer (multiple "print passes")
;	--------------------
;
PRINT_ONE_PXL_LINE PROC	NEAR
	PUSH	SI		; Save buffer location.
	PUSH	CX
	PUSH	BX
	PUSH	AX

	mov	cx,[bx]		; Get # of columns in this "pass"

	cmp	cx,MAX_BUFLEN
	jle	PRN1_PROCESS
	mov	MAX_BUFLEN,cx	; This buffer is longer, save this count.
PRN1_PROCESS:

	mov	al,BUF_NUM
	cmp	al,PIN24-1	; if 3 passes, fill extra buffers 0..1
	jge	PRN1_SEND	; Last pass? leave in place & send all

	inc	al		; Extra buffer n (=0..2) at ML_BK + BUFLEN*(n+1)
	mov	BUF_NUM,al	; Increment for next time
	mov	si,bx		; Address of ML_BK's length word
	push	di
	 mov	di,si		; Copy destination = extra buffer
	 push	cx
	  mov	cl,al
	  xor	ch,ch
PRN1_ADDR_LOOP:			; Find address of buffer N
	  add	di,BUFLEN
	  loop	PRN1_ADDR_LOOP
	 pop	cx		; CX = # of data bytes
	 add	cx,2		; Data bytes + 2 bytes length
	 rep	movsb		; Copy data to extra buffer
	pop	di

	xor	ax,ax		; Show no error
	jmp	short PRN1_XIT	; We saved the data; return w/o sending yet
	
PRN1_SEND:

	mov	cx,MAX_BUFLEN	; Maximum column count
	jcxz	PRN1_LFONLY	; 0 columns: send CR/LF only

 IF (PRN_GR AND GR_COLS) EQ GR_COLS ; Send # of columns
	call	PRINT_ONE_START_GFX
 ELSE
	; Data length = # of print passes * # of columns
	push	cx
	 push	bx
	  mov	bx,cx		; BX = # of columns
	  add	cx,bx
  IF PIN24 GT 2
	  add	cx,bx
   IF PIN24 GT 3
	  add	cx,bx
   ENDIF
  ENDIF
  IFDEF GR_LENOFF
	  add	cx,GR_LENOFF	; Offset to length (ProPrinter only?)
  ENDIF
	  call	PRINT_ONE_START_GFX
	 pop	bx
	pop	cx
 ENDIF
	jnz	PRN1_XIT0
	add	bx,2		; Go to first data byte
	; Now send one byte from each:
	;	extra buffer 0
	;	  "	"    1 (if PIN24>2)
	;	  "	"    2 (if PIN24>3)
	;	standard buffer (addr in BX)
	; and then repeat.
PRN1_SENDLOOP:
	mov	al,[bx+BUFLEN]
	call	PRINT_ONE_CHAR
	jnz	PRN1_XIT0
 IF PIN24 GT 2
	mov	al,[bx+BUFLEN*2]
	call	PRINT_ONE_CHAR
	jnz	PRN1_XIT0
  IF PIN24 GT 3
	mov	al,[bx+BUFLEN*3]
	call	PRINT_ONE_CHAR
	jnz	PRN1_XIT0
  ENDIF
 ENDIF	
	mov	al,[bx]
	call	PRINT_ONE_CHAR
	jnz	PRN1_XIT0

	inc	bx
	loop	PRN1_SENDLOOP

PRN1_LFONLY:
	MOV	SI,OFFSET CRLF
	CALL	PRINT_IT
	jnz	PRN1_XIT0

PRN1_XITOK:
	xor	ax,ax		; 'Zero' indicates no error
PRN1_XIT0:
	; Don't destory flags, as 'Z' flag tells whether error
	pushf
	push	di
	 mov	di,offset ml_buf ; Zero out the data buffers
	 mov	ax,0
	 mov	cx,(BUFLEN*PIN24)/2 ; Clear all buffers
	 cld			; Ensure forward direction
	 rep	stosw		; (CX=0 at end)
	pop	di
	popf

	; Zero maximum column count & Buffer #
	mov	MAX_BUFLEN,ax
	mov	BUF_NUM,al

PRN1_XIT:
	POP	AX
	POP	BX
	POP	CX
	POP	SI
	RETN
PRINT_ONE_PXL_LINE ENDP
ELSE
;
; ************************************
;
;	PRINT_ONE_PXL_LINE
;
;	Sends a line of pixels to the printer.
;
;	Entry:  DS:BX	Points to data buffer to be sent to printer
;
;	Exit:	'Zero' if printed OK, not 'Zero' if error
;
;	Uses:	Preserves all
;
;	--------------------
;	Standard (Laser, Inkjet, 8 pin Matrix) logic
;	--------------------
;
PRINT_ONE_PXL_LINE PROC	NEAR
	PUSH	SI		; Save buffer location.
	PUSH	CX
	PUSH	BX
	PUSH	AX

	xor	ax,ax		; Assume no error

 IF (PRN_GR AND GR_BITS) NE GR_FIX ; If not fixed # of columns
	mov	cx,[bx]		; Get # of bytes to print
	JCXZ	PRN1_XIT	; Nothing to do, just exit.
 ELSE
	; Fixed length - retrieve length from print command
	mov	si,offset gmodes ; Offset to graphics mode table
	xor	ah,ah
	mov	al,gmode	; Graphics mode
	sal	al,1		;  x8 (each entry is 8 bytes long)
	sal	al,1
	sal	al,1
	add	si,ax		; Point SI at gmodes[gmode]
	push	bx
	 mov	bx,[si+2]	; Point DX at length[gmodes[gmode]]
	 mov	cx,[bx]		; Get length from print command
	pop	bx
 ENDIF

 IF (PRN_GR AND GR_NOT) EQ GR_NOT
	push	cx
	push	bx
	add	bx,2		; first data byte
	mov	al,0FFH
prn1_notloop:
	xor	byte ptr [bx],al ; Invert each byte in the buffer
	inc	bx
	loop	prn1_notloop
	pop	bx
	pop	cx
 ENDIF
 IF (PRN_GR AND GR_RUN) EQ GR_RUN
	CALL	RUN_COMPRESS	; Compress output buffer
	CALL	PRINT_IT	; Send sequence to select normal/runlength
 ENDIF
	CALL	PRINT_ONE_START_GFX
	jnz	prn1_xit	; Stop if error

 IFDEF QUOTE_FLAG
	mov	al,1		; In graphics -- must quote Ctrl-C's
	mov	quote_flag,al	; (if GR_QUO) or set bit (if GR_BIT)
 ENDIF

	MOV	SI,BX		; Point to print buffer
	MOV	[SI],CX
	CALL	PRINT_IT	; Send graphics info
	jnz	prn1_xit	; Error -- exit.

 IFDEF QUOTE_FLAG
	xor	al,al
	mov	quote_flag,al	; Don't quote Ctrl-C's -- not in graphics.
 ENDIF
	; Send sequence to end graphics mode and home carriage, as applicable.
	; Prepare to select next ribbon or color plane on color printers.
	mov	si,offset gr_off
	call	print_it

PRN1_XIT:
	POP	AX
	POP	BX
	POP	CX
	POP	SI
	RETN
PRINT_ONE_PXL_LINE ENDP
ENDIF	; ifdef pin24
;
; ************************************
;
;	PRINT_ONE_START_GFX
;
;	Sends the command to put the printer into the selected graphics
;	mode (sending a command length if required).
;
;	Entry:  CX	Number to be sent to printer as command length
;			(may be # of data bytes or # of columns)
;
;	Exit:	'Zero' if command sent properly, 'NonZero' otherwise
;		CX	Actual length of buffer to send
;
;	Uses:	AX, SI
;
PRINT_ONE_START_GFX PROC NEAR

	mov	si,offset gmodes ; Offset to graphics mode table
	xor	ah,ah
	mov	al,gmode	; Graphics mode
	sal	al,1		;  x8 (each entry is 8 bytes long)
	sal	al,1
	sal	al,1
	add	si,ax		; Point SI at gmodes[gmode]

IF (((PRN_GR AND GR_BITS) EQ GR_DEC) OR ((PRN_GR AND GR_BITS) EQ GR_DE2))
	; NEC 8023/ C. Itoh / HP LaserJet require 4-char 'decimal' numbers
	; Canon LBP requires the count be sent _twice_.
	push	bx
	push	cx
	 mov	ax,cx		; AX = length
	 push	dx
	 push	di
	  mov	di,[si+2]	; Point DI at length buffer
	  mov	cx,GR_DEC_DIGITS ; usually 3 or 4 characters
	  add	di,cx		; DI = last char
	  dec	di
	  mov	bx,10		; for decimal division
p1p_lp:	  xor	dx,dx
	  div	bx		; DX = remainder (this digit)
	  add	dl,'0'
	  mov	[di],dl
	  dec	di
	  loop	p1p_lp		; Do for each digit
IF (PRN_GR AND GR_BITS) EQ GR_DE2
	  push	si
	   mov	si,[si+2]	; Point SI at length buffer
	   mov	di,offset PMODE_LENALL
	   mov	cx,GR_DEC_DIGITS
	   cld
;	   push es		; Use DS as string destination
;	   mov	dx,ds
;	   mov	es,dx
	   rep	movsb		; Copy the count into the second place
;	   pop	es
	  pop	si
ENDIF
	 pop	di
	 pop	dx
	pop	cx
	pop	bx
ENDIF
IF (PRN_GR AND GR_BITS) EQ GR_BIN
	push	bx
	 mov	bx,[si+2]	; Point DX at length[gmodes[gmode]]
	 mov	[bx],cx		; Save length in print command
	pop	bx
ENDIF
IF (PRN_GR AND GR_BITS) EQ GR_BIX
	push	bx
	 mov	bx,[si+2]	; Point DX at length[gmodes[gmode]]
	 mov	byte ptr [bx],ch ; Save length, MSB first
	 inc	bx
	 mov	byte ptr [bx],cl
	pop	bx
ENDIF
	mov	si,[si]		; Point SI at print command
	call	print_it	; Set graphics mode.
	RETN
PRINT_ONE_START_GFX ENDP

IF (PRN_GR AND GR_RUN)
;
; ************************************
;
;	RUN_COMPRESS
;
;	Pair-type Run-length encoding (count,value,count,value...)
;	for HP printers.  Uses Black buffer to build pair encoded data.
;
;	Entry:	BX points to buffer
;		CX has length of buffer
;	Exit:	BX points to buffer to use (either original BX, or Black area)
;		CX is length
;		SI points to compression-mode command string
;	Uses:	AX
;
	PUBLIC RUN_COMPRESS
RUN_COMPRESS PROC NEAR

;	INT	3

	push	bx
	push	cx
	push	dx
	mov	si,bx		; Get address of buffer
	inc	si		; skip over length word
	inc	si
	mov	bx,offset ML_ENCODE+2 ; Place to build data (skip over length word)
;  DS:[SI]	source buffer
;  BX		output buffer
;  DX		count for this pair
;  AH		previous byte
;  AL		current byte
	xor	dx,dx		; init DX = 0
	mov	ah,[si]		; Get first byte
	inc	si

print_one_rl_loop:
	lodsb			; Get byte in AX, inc SI
	inc	dx
	cmp	ah,al
	loope	print_one_rl_loop ; decrement CX and loop if bytes same
	jcxz	print_one_rl_done ; all done? quit.
	call	print_one_rl_save
	jmp	print_one_rl_loop

print_one_rl_done:
	call	near ptr print_one_rl_save ; Save final pair, if possible
	pop	dx
	pop	cx		; original count

;	INT	3		; [[[ !!! *** DEBUG *** !!! ]]]

	sub	bx,offset ML_ENCODE+2	; BX = # of bytes created
	mov	word ptr ML_ENCODE,bx	; Save new length

	mov	si,offset xmit_norm ; pointer to command for normal mode
IFDEF xmit_nochg
	; If printer allows us to stay in previous compression mode
	; the length of the command to CHANGE modes is considered too.
	cmp	si,last_compression ; Is current mode uncompressed?
	jne	print_one_rl_comp0
	add	bx,word ptr xmit_runl ; Yes. Compressed size will include
				; length of set-to-runlength command.
	jmp	short print_one_rl_comp1
print_one_rl_comp0:
	add	cx,[si]		; Add length of set-to-normal to uncompressed.
print_one_rl_comp1:
ENDIF

	cmp	bx,cx		; is compressed size {BX} < uncompressed {CX}?
	pop	bx		; BX = original buffer ptr
	jg	print_one_rl_go ; compressed bigger -- use uncompressed
	mov	bx,offset ML_ENCODE ; use compressed buffer
	mov	si,offset xmit_runl ; Set printer into runlength mode
print_one_rl_go:
	mov	cx,[bx]		; get length of whichever buffer
IFDEF xmit_nochg
	; If printer allows us to stay in previous compression mode
	cmp	si,last_compression ; determine whether to send compression cmd
	je	print_one_rl_gsame  ; Same - don't send cmd
	mov	last_compression,si ; Different - send cmd & remember
	retn
print_one_rl_gsame:
	mov	si,offset xmit_nochg ; Don't send cmd
	retn
ELSE
	retn			; always send compression string
ENDIF

print_one_rl_save:
	cmp	bx,OFFSET ML_ENCODE+BUFLEN ; gone too far?
	jge	print_one_rl_s1 ; yes -- can't save

	cmp	dx,GR_RUN_MAX	; too many for one pair?
	jg	print_one_rl_s2

IF (GR_RUN_OFFSET EQ -1)
	dec	dx		; Count=0 makes 1 byte
ELSE
 IF (GR_RUN_OFFSET EQ -257)
	neg	dx		; DX = 256 - DX
	inc	dx		; DX = 257 - DX
	; NOTE: count=2 is rendered as 255; count=3 becomes 254;
	; but count=1 becomes zero (!) -- just the way that
	; PostScript wants it.
 ELSE
  IF (GR_RUN_OFFSET NE 0)
	add	dx,GR_RUN_OFFSET ; add other offset
  ENDIF
 ENDIF
ENDIF
	mov	[bx],dl		; save count
	inc	bx
	mov	[bx],ah		; save value
	inc	bx		; another pair
	mov	ah,al		; save current byte in AH
print_one_rl_s1:
	xor	dx,dx		; DX = 0
	retn	

print_one_rl_s2:
	mov	byte ptr [bx],GR_RUN_MAX+GR_RUN_OFFSET	; save max count
	inc	bx
	mov	[bx],ah		; save previous value
	inc	bx
	sub	dx,GR_RUN_MAX	; count down
	jmp	print_one_rl_save ; rest of run

RUN_COMPRESS ENDP
ENDIF

	SUBTTL	EGAD Printer Driver (Subroutines: DO_PIXELS)
;
; ************************************
;
;	DO_PIXELS
;
;	Read one screen line, and fill the printer buffer.
;
;	Input:
;		BX points to LINES_xxx_PATTERN
;		SCREEN_LINE (used by READ_DOT)
;		-or- SCREEN_COL (if ROTATE is true)
;		OUT_BITS (used by SET_DOT)
;	Uses:
;		SCREEN_COL
;		-or- SCREEN_LINE (if ROTATE is true)
;
;	Preserves: BX, CX, DX, SI, BP
;
DO_PIXELS PROC NEAR
	push	si
	push	dx
	push	cx
	push	bx

IF (PRN_BYTE EQ BYTE_HORIZ)
	push	bp

	mov	bp,bx		; BP points to LINES_xxx_PATTERN
	xor	dh,dh
	mov	dl,[bx]		; DX = # of pixel patterns
	inc	bp		; Point BP at first byte pattern
	xor	di,di		; Use first byte pattern (DI=0)
	mov	al,ds:[bp][di]	; Get byte pattern
	mov	OUT_BITS,al	; Save pattern in OUT_BITS
ENDIF

	mov	bx,graph_lmarg	; Graphics left margin
	mov	out_loc,bx	; First byte of printer buffer

	xor	bx,bx		; BX = 0
	cmp	rotate,bl	; Is rotate=0?
	je	do_pix_0

	mov	si,offset SCREEN_LINE
	mov	[si],bx		; Set to 0th line in current column
	add	bx,cfg.y1	; align	to screen pixel
	mov	cx,cfg.y2
	jmp	short do_pix_loop

do_pix_0:
	mov	si,offset SCREEN_COL
	mov	[si],bx		; Set to 0th pixel on line
	add	bx,cfg.x1	; align	to screen pixel
	mov	cx,cfg.x2

; Get pixel's x	coordinate.  SCREEN_COL always ranges from 0..319
; or 0..639.  (Print pixel number.)  READ_DOT will change this to range from
; cfg.x1 to cfg.x1+319 (or cfg.x1+639) (Screen pixel number).

;	BX is SCREEN_COL + CFG.X1, which counts up from CFG.X1 to CFG.X2.
;	CX is CFG.X2.
;	SI points to the variable we increment for each dot:
;		SCREEN_COL (non-rotate) or SCREEN_LINE (rotate)
; For Laser printers:
;	BP points to the first byte pattern
;	DX is the number of screen pixels per printer byte
;	DI is the byte pattern number to use (0..n) for Laser printers

do_pix_loop:

	cmp	bx,cx		; Don't print this pixel if
	jg	do_pix_done	; SCREEN_COL + CFG.X1 > CFG.X2
				; (SCREEN_LINE + CFG.Y1 > CFG.Y2 if Rotating)

	CALL	SET_DOT		; Plot pixel(s) in print buffer(s).

 IF (PRN_BYTE EQ BYTE_HORIZ)
	inc	di		; Go to next bit pattern
	cmp	di,dx		; Is this pixel in next printer byte column?
	jl	do_pix_1
	xor	di,di		; Back to first pattern
	inc	out_loc		; Offset in buffer goes up now, as we
				; must pack bits into bytes.
do_pix_1:
	mov	al,ds:[bp][di]	; Get byte pattern
	mov	OUT_BITS,al	; Save pattern in OUT_BITS
 ENDIF

 if (prn_byte eq byte_vert)
  ; On Dot Matrix printers, change Dithers each Pixel.
  ; The dither table is assumed to contain dithers of the form:
  ;	a b c d a b c d
  ; e.g., if the DITHER_WIDTH is 8, then offsets 0..3 into the
  ; table will be used.  This will produce the same dithers
  ; regardless of the prn_repeat factor.  For example if the
  ; prn_repeat is 3, the following patterns will be used:
  ;	Offset= 0	a b c
  ;		3	d a b
  ;		2	c d a
  ;		1	b c d	(then offset 0 again.)
  ; Producing patterns: a b c d a b c d a b c d ...
  ; (code uses DI)
	mov	ax,word ptr prn_repeat
	add	out_loc,ax		; Move to place to put next scrn pixels
	add	al,dither_num		; Add repeat to current dither number
	and	al,(DITHER_WIDTH/2)-1	; Map to 0..1, 0..3, etc.
	mov	dither_num,al

	mov	di,offset dither_table	; Base of table 0
	add	di,ax			; Index into each element of table
	mov	dither_ptr,di		; Save as new base
 ENDIF
	inc	bx		; Go to next printer pixel.
	inc	word ptr [si]	; Increment SCREEN_LINE or SCREEN_COL.
	jmp	do_pix_loop

do_pix_done:

IF (PRN_BYTE EQ BYTE_HORIZ)
	pop	bp
ENDIF
	pop	bx
	pop	cx
	pop	dx
	pop	si

	RETN

DO_PIXELS ENDP
	SUBTTL	EGAD Printer Driver (Subroutines: PRINT_ONE_CHAR)
;
; ************************************
;
;	PRINT_ONE_CHAR
;
;	Prints a single character.
;
;	Entry:	AL contains character to be printed
;	Exit:	AH = 0 if no error, <>0 if error
;		'Z' flag true if no error, false if error
;	Uses:	AH, DX
;
PRINT_ONE_CHAR PROC NEAR
	MOV	DX,cfg.prn_un	; Get printer number
	OR	DH,DH		; Check sign bit
	JS	PO_SER		; If set, negative #'s indicate Serial I/O
; Parallel Output: (port # 1..3 in DX)
	DEC	DX		; (make 0 based)
	JL	PO_Z		; If PRINTER=0, use Z-100 output
	XOR	AH,AH		; Print character (AH=0)
	INT	17H
	AND	AH,09h		; I/O Error or Timeout
	RET
; Serial Output: (port # is -DX, -1..-4)
PO_SER:	NEG	DX		; Get port #
	DEC	DX		; (make 0 based)
	MOV	AH,1		; Print character
	INT	14H
	AND	AH,80H		; Time-out error?
	RET
; Z-100 BIOS Printer Output
PO_Z:	
	CALL	FAR PTR ZBIOS_PRINT
	XOR	AX,AX		; Show no error!
	RET
PRINT_ONE_CHAR ENDP

	SUBTTL	EGAD Printer Driver (Subroutines: PRINT_IT)
;
; ************************************
;
;	PRINT_IT
;
;	Entry:  DS:[SI] points to data to be printed.
;
;	Exit:	Selected data has been printed
;		If OK, ERRCOD is zero, else one.
;
; First	two bytes of string contain the	character count.
; All registers	preserved on return.
PRINT_IT PROC NEAR
	PUSH	DI
	PUSH	SI
	PUSH	DX
	PUSH	CX
	PUSH	AX
	MOV	WORD PTR ERRCOD,0 ; Reset print	error
	LODSW		; Get character count
	MOV	CX,AX	;   in CX
	JCXZ	PI_XIT	; Nothing to do: Exit
PI_LOOP:
	LODSB		; Get character
IF (PRN_GR AND GR_BITS) EQ GR_QUO
	cmp	al,GR_QUOTE	; Quote this character?
	jne	pi_1a
	cmp	quote_flag,0
	jz	pi_1a
	call	PRINT_ONE_CHAR ; send the ^C twice
	jnz	PI_ERR	; Abort if error
pi_1a:
ENDIF
IF (PRN_GR AND GR_BITS) EQ GR_BIT
	cmp	quote_flag,0
	jz	pi_1a		; Not in graphics -- don't set bit
	OR	al,GR_BIT_VAL	; Yes, in graphics -- set bit so printer knows
pi_1a:
ENDIF
IF (PRN_GR AND GR_XHEX) EQ GR_XHEX
	cmp	quote_flag,0
	jz	pi_1b		; Send through a normal character
	; For Postscript printers, encode binary to ASCII hex char's
	PUSH	AX
	SHR	AL,1		; Send high nybble
	SHR	AL,1
	SHR	AL,1
	SHR	AL,1
	ADD	AL,'0'
	CMP	AL,'9'
	JLE	pi_1a
	ADD	AL,'A'-10-'0'
pi_1a:	CALL	PRINT_ONE_CHAR
	POP	AX		; must pop before jumping!
	JNZ	PI_ERR

	AND	AL,15		; Send low nybble
	ADD	AL,'0'
	CMP	AL,'9'
	JLE	pi_1b
	ADD	AL,'A'-10-'0'
pi_1b:
ENDIF
	CALL	PRINT_ONE_CHAR
	JNZ	PI_ERR	; Abort if error
	LOOP	PI_LOOP	; Send all characters
	XOR	AH,AH	; Show successful completion
PI_XIT:	TEST	AH,0FFH	; Set zero flag	for error return
	POP	AX
	POP	CX
	POP	DX
	POP	SI
	POP	DI
	RETN
PI_ERR:	MOV	BYTE PTR ERRCOD,1 ; Show error in PRINT_IT
	JMP	PI_XIT
PRINT_IT ENDP
	SUBTTL	EGAD Printer Driver (Subroutines: SET_DOT)
;
; ************************************
;
;	SET_DOT
;
;	Entry:
;	  OUT_LOC	Byte offset into print buffer
;	  OUT_BITS	which of the printer pins to fire for this
;			horizontal line.  It's a bit mask, so we can
;			set several pixels at once.
;			(On a laser printer, which bits to set in the
;			current horizontal byte.)
;
;	Exit:	 One pixel of each print color in all buffers.
;	Uses:	 AX, DI
;
SET_DOT PROC NEAR
	PUSH	BP
	PUSH	BX
	PUSH	CX
	PUSH	DX
	PUSH	SI
 IF (PRN_BYTE EQ BYTE_HORIZ)
	PUSH	DI		; laser drivers uses DI in DO_PIXELS
 ENDIF
	CALL	READ_DOT	; Returns color # in AX

	mov	di,BUF_START	; Data position of first pixel in current buffer
	add	di,out_loc	; Offset into buffer

	MOV	BX,OFFSET COLOR_TABLE ; Pixel value is offset to translation
	XLAT			; Translate screen color to abstract color (AL)
 IFDEF COLOR_XLAT
	; Translate abstract color numbers (--yymmcc) to physical printer #'s
	; -- WL 10/01/93
	mov	bx,offset color_xlat
	xlat
 ENDIF
 IF DITHER_LEVELS GT 200
	; Printer handles dithering.
	; Simply pass physical color number to the printer.
	; -- WL 10/01/93
	mov	[di],al		; Store in buffer
	jmp	set_dot_exits
 ENDIF

	mov	si,dither_ptr	; Pointer to current dither table
	mov	dl,byte ptr OUT_BITS ; Bits to set in the given byte
	xor	dh,dh		; Assume set just one buffer byte for each pixel
	xor	bx,bx
	mov	ah,byte ptr prn_repeat	; # Printer pixels per screen pixel

	; AH = video_switch (usually 1..3) (always 1 on Laser printer)
	; AL = Printer color #
	; DI = Pointer to print buffer
	; SI = Pointer to Dither Table
	; DL = nth bit(s) to OR into byte in buffer,
	; DH = copy of DL (if prn_repeat > 1)
	; BL = copy of DL (if prn_repeat > 2)
	; BH = copy of DL (if prn_repeat > 3)

 if (prn_byte eq byte_vert)
	dec	ah		; Is AH=1?
	jz	sl_y		; 1:1 correspondence, just set 1 bit.
	mov	dh,dl		; Set both bytes.

	dec	ah		; Is AH=2?
	jz	sl_y
	mov	bl,dl		; 3:1 factor.

	dec	ah		; Is AH=3?
	jz	sl_y
	mov	bh,dl		; 4:1 factor.
 ENDIF

SL_Y:

	; CL = # of bits to shift printer color right
	; CH = mask for current 2 color bits
	; BP = # of buffers (1..4) to do.

	mov	bp,NBUFFERS	; Number of buffers
	
 IF DITHER_LEVELS EQ 2
	mov	ch,00000001b	; ON/OFF only
 ENDIF
 IF DITHER_LEVELS EQ 4		; 4 color/gray scale levels
	mov	ch,00000011b	; 2 bits for color
				; (Yellow on color printers, Black for B&W)
 ENDIF
 IF DITHER_LEVELS EQ 8
	mov	ch,00000111b	; 8 levels of gray
 ENDIF
 IF DITHER_LEVELS EQ 16
	mov	ch,00001111b	; 16 levels of gray
 ENDIF

	xor	cl,cl		; color numbers start at 0
	xor	ah,ah		; AH=0

SET_DOTS:
	push	dx
	push	bx
	push	ax

	push	di		; save DI
	mov	di,bx		; put BX into DI
	and	al,ch		; Look at current two color bits only
	shr	ax,cl		; Make them bottom bits.

 ; If dither_width = 1, no action.
 IF DITHER_WIDTH EQ 2
	shl	ax,1		; Multiply by 2 (bytes/entry)
 ENDIF
 IF DITHER_WIDTH EQ 4
	shl	ax,1		; Multiply by 4 (bytes/entry)
	shl	ax,1
 ENDIF
 IF DITHER_WIDTH EQ 8
	shl	ax,1		; Multiply by 8 (bytes/entry)
	shl	ax,1
	shl	ax,1
 ENDIF
	mov	bx,ax
	and	dx,[si][bx]	; Use Dither Table to get rid of selected #
 IF DITHER_WIDTH GT 2
	and	di,[si][bx+2]	; of bits from the OR-masks.
 ENDIF
	mov	bx,di		; restore bits in BX
	pop	di		; restore DI

	OR	word ptr [DI],dx ; Set pixel(s)
 IF DITHER_WIDTH GT 2
	OR	word ptr [DI+2],bx ; Set more pixels.
 ENDIF

	pop	ax
	pop	bx
	pop	dx

 IFDEF ML_YW ; if separate color buffers
	inc	cl		; Next bit or bits
	shl	ch,1
  IF DITHER_LEVELS EQ 4
	inc	cl		; Two bits per color #
	shl	ch,1		; Move mask over 2 bits
  ENDIF
	add	di,buflen	; Go to next buffer

	dec	bp		; Done with one buffer.
	jg	set_dots	; Do more?
 ENDIF

set_dot_exits:
 IF (PRN_BYTE EQ BYTE_HORIZ)
	POP	DI
 ENDIF
	POP	SI
	POP	DX
	POP	CX
	POP	BX
	POP	BP
	RETN
SET_DOT ENDP
	SUBTTL	EGAD Printer Driver (Subroutines: READ_DOT)
;
; ************************************
;
;	READ_DOT
;
;	Reads 'current' pixel from the screen.  Does all necessary
;	pre-processing (VGA division, etc.)  Perform rotation of
;	axes if selected.
;
;	Entry:
;	  SCREEN_LINE,
;	  SCREEN_COL	contain index into crop box (offsets from
;			CFG.X1, CFG.Y1).
;
;	Exit:	AL = screen color number
;
;	Uses:	AX,BX,CX,DX
;
READ_DOT PROC NEAR
	MOV	DX,SCREEN_LINE
	MOV	CX,SCREEN_COL

	; ROTATE=	CX,SCREEN_COL range	DX,SCREEN_LINE range
	;    0		0..cfg.x2-cfg.x1		cfg.y1..cfg.y2
	;    1		cfg.x1..cfg.x2*		0..cfg.y2-cfg.y1
	;		^ This makes for a "mirrored" printout,
	;		  so we have to subtract each value from cfg.x2.

	cmp	rotate,0	
	je	read_dot_p

; READ_DOT_L:		; Read landscape
	mov	ax,cfg.x2
	sub	ax,cx		; Rotating: make CX 0..cfg.x2-cfg.x1,
	mov	cx,ax		;   in 'backwards' order.
	jmp	short read_dot_PL

READ_DOT_P:		; Read portrait
	sub	dx,cfg.y1	; make DX 0..cfg.y2-cfg.y1

READ_DOT_PL:		; Read (both orientations)
	mov	al,byte ptr cfg.enlarge

	or	al,al		; Is it zero? (i.e., normal size?)
	jnz	rd_enlarge	; No - process other enlargement

READ_DOT_GO:
	add	cx,cfg.x1	; CX,DX are offsets; make them absolute.
	add	dx,cfg.y1

	mov	ax,cx		; (AX,BX) is point to read
	mov	bx,dx
	call	near ptr readpt	; Call pixel-read routine in VIDEO_PC or VIDEO_Z
	and	al,ah		; Keep only appropriate bits. -- wl 11/09/94

	INC	AH		; If AH was 255, this indicates 256-color mode.
	JZ	rd_256c		; (incrementing it sets AH to zero)

READ_DOT_EXIT:

	; AND	AL,0FH		; Ensure color 0..15 (index into PALETTE)
	RETN

RD_ENLARGE:
	dec	al
	jz	rd_x2		; * 2

	dec	al
	jz	rd_x3		; * 3

	dec	al
	jz	rd_x4
	; Assume values > 4 indicate 1 1/2x.
	jmp	short rd_x5

rd_x4:	shr	cx,1
	shr	dx,1

rd_x2:	shr	cx,1
	shr	dx,1
	jmp	read_dot_go

; 3x enlargement
rd_x3:	mov	bx,3		; Divide by 3
	; Divide X coordinate by 3
	push	dx
	 sub	dx,dx		; DX = 0
	 mov	ax,cx		; 
	 div	bx		; AX = DX:AX / BX ( X / 3 )
	 mov	cx,ax		; save in CX
	pop	ax		; Put new Y into AX
	sub	dx,dx
	div	bx		; AX = DX:AX / BX ( Y / 3 )
	mov	dx,ax		; Put new X into AX
	jmp	read_dot_go

; 5 = 1.5x enlargement
rd_x5:	add	cx,cx		; Multiply both CX...
	add	dx,dx		; ...and DX by 2...
	jmp	rd_x3		; then divide by 3

rd_256c:
; In 256-color VGA mode, divide color number, then take modulo
; Look for standard divisors, and do shifting if we can
	MOV	AH,byte ptr cfg.vga_divisor
	CMP	AH,1		; if =1, nothing to do
	JZ	read_dot_exit
	CMP	AH,2
	JZ	RD_D2
	CMP	AH,4
	JZ	RD_D4
	CMP	AH,8
	JZ	RD_D8
	CMP	AH,16
	JZ	RD_D16
	CMP	AH,32
	JZ	RD_D32

; Non-standard divisor - conduct division (much slower!)
	XOR	AH,AH
	DIV	byte ptr cfg.vga_divisor
	JMP	read_dot_exit

rd_d32:	SHR	AL,1
rd_d16:	SHR	AL,1
rd_d8:	SHR	AL,1
rd_d4:	SHR	AL,1
rd_d2:	SHR	AL,1
	jmp	read_dot_exit

READ_DOT ENDP

_TEXT	ENDS
	END	START
