;
; MANUFACT.INC
;
; Lists Equates for printer manufacturer and model types
;
;
; ******************************
; Printer type codes, TYP_*
; Note: These are bit-fields and can be OR'ed together.
; ******************************

TYP_COLOR	=  1		  ; Any color printer
TYP_CTEXT	=  2		  ; Color text available
				  ; (Otherwise text attributes)
TYP_LASER	=  4		  ; Lasers, incl. HP and PostScript
TYP_MATRIX	=  8		  ; all matrix printers
TYP_MAT24	= 16 + TYP_MATRIX ; 24-pin matrix only
TYP_INKJET	= 32		  ; Inkjets not otherwise compatible
TYP_RESERVED2	= 64		  ; **(reserved)

; Color combinations

TYP_COLORLASER  = TYP_COLOR + TYP_LASER
TYP_COLORMATRIX = TYP_COLOR + TYP_MATRIX + TYP_CTEXT
TYP_COLORMAT24  = TYP_COLOR + TYP_MAT24 + TYP_CTEXT
TYP_COLORINKJET = TYP_COLOR + TYP_INKJET
TYP_COLORINKJET2 = TYP_COLOR + TYP_INKJET + TYP_CTEXT

; Manufacturer's codes, MFG_*

MFG_HP		=  0
MFG_EPSON	=  1	; Includes clones (and Star Gemini)
MFG_IBM		=  2	; non-Epson compatibles
MFG_TOSHIBA	=  3
MFG_NEC		=  4
MFG_DP		=  5
MFG_OKI		=  6
MFG_MPI		=  7
MFG_PCPI	=  8
MFG_XEROX	=  9
MFG_TANDY	= 10
MFG_ADOBE	= 11	; any PostScript printer
MFG_CANON	= 12

; Model codes, MDL_*

; HP models:
;   "PCL-3" (no IBM PC character set)
MDL_HPPCL3	= 0
MDL_HPLJ	= 0
;   "PCL-4" (with IBM PC character set)
MDL_HPPCL4	= 1
MDL_HPLJ2	= 1
MDL_HPQJ	= 2	; QuietJet

;   "PCL-5" (with data compression)
MDL_HPPCL5	= 3
MDL_HPLJ3	= 3
MDL_HPLJ4	= 4
MDL_HPPJ	= 5	; PaintJet
MDL_HPDJ500	= 6	; DeskJet 500 (or 500C if TYP_COLOR)
MDL_HPDJ550	= 7	; DeskJet 550C

; Epson models
MDL_EPMX	= 0	; no 240 dpi graphics
MDL_EPFX	= 1	; with 240 dpi graphics
MDL_EPST	= 2	; Star Micronics 240 dpi codes
MDL_EPLQ	= 3	; Epson LQ 24-pin

; IBM models
MDL_IBMPP	= 0	; ProPrinter II (9-pin); and 9-pin emulation
			; for ProPrinter X24 with AGM *OFF*
MDL_IBMX24	= 1	; ProPrinter X24 (24-pin mode, AGM *OFF*)
			; *NOTE* With AGM *ON*, ProPrinter acts like
			; Epson compatible for our purposes.
			; [except: NO ITALICS...]
MDL_IBM3852	= 2	; Color Inkjet 3852

; Toshiba models
MDL_TOSHP3X1	= 0	; 24-pin models

; NEC models
MDL_NEC8023	= 0	; NEC-8023, C.Itoh 8510, Apple ImageWriter I

; DataProducts (Integral Data Systems)
MDL_DPTIGER	= 0	; Paper Tiger (B&W)
MDL_DP8050	= 1	; Dataproducts 8050 color printer

; Okidata
MDL_OKI82	= 0	; Okidata 82/83/84/92/93/192/193
MDL_OKI83	= 1
MDL_OKI84	= 2
MDL_OKI92	= 3
MDL_OKI93	= 4
MDL_OKI292	= 5	; Color, quasi-Epson compatible

; MPI
MDL_MPI99	= 0	; MPI-99G,MPI-150G

; PCPI
MDL_PCPI2000	= 0	; LaserImage 2000

; Xerox
MDL_XEROX4020	= 0

; Tandy
MDL_RSDMP105	= 0	; DMP-105 matrix
MDL_RSCGP220	= 1	; CGP-220 inkjet

; PostScript
MDL_LEVEL1	= 1	;
MDL_LEVEL2	= 2	; PostScript extensions needed for color

; Canon
MDL_LBP4	= 0	; LBP-4, 8, 4LT, etc.

; ******************************
;	Character set types
; ******************************
CHRSET_ASCII7	= 0	; Just ASCII codes
CHRSET_PC8	= 1	; IBM PC character set
CHRSET_ROMAN8	= 2	; like in original HP LaserJet
CHRSET_ADOBE	= 3	; PostScript set
CHRSET_OTHER	= 4	; Other character set

; ******************************
;	Printer characterstics
; ******************************
BYTE_VERT	= 0	; Vertical bytes, like on matrix printers
BYTE_HORIZ	= 1	; Horizontal bytes, like on laser printer

; ******************************
; Graphics-mode data-descriptor encoding type
; (shown with example printer)
; ******************************

GR_BITS EQU	7		; 3 bits describing graphics type

GR_BIN	EQU	0		; 2 binary bytes, LSB first (Epson)
GR_BIX	EQU	1		; 2 binary bytes, MSB first
GR_DEC	EQU	2		; 4 decimal ASCII digits (NEC, HP)
GR_QUO	EQU	3		; Escape-character quoting (Okidata)
GR_BIT	EQU	4		; specific bit set indicates graphics (MPI)
GR_FIX	EQU	5		; fixed length (word at PMODE_LENx).
GR_DE2	EQU	6		; Two decimal (GR_DEC) counts required

; These values may be OR'ed with the above values:
GR_XHEX	EQU	16		; Expand binary gfx data to hex char's
GR_COLS	EQU	32		; Data transfer command uses print COLUMNS
				; not byte count. (e.g., Toshiba 24-pin printers)
GR_RUN	EQU	64		; Run-length pairs ((count,value)...)
GR_NOT	EQU	128		; Invert binary values before sending

